#include <windows.h>


#define BMP_BUFFERSIZE      51200

#define BMP_FLIP            4
#define BMP_DONT_FLIP       0
#define BMP_INVERT          5
#define BMP_DONT_INVERT     0
#define BMP_HEADER          6
#define BMP_NO_HEADER       0
#define BMP_PALETTE         7
#define BMP_NO_PALETTE      0
#define BMP_ALIGN           8
#define BMP_DONT_ALIGN      0


#define BMP_FILE_ERROR      1
#define BMP_HEADER_ERROR    2
#define BMP_BITCOUNT_ERROR  3

#define BMP_ERROR(err)      lbmp->errorcode = err; return

#define makeCol15(r,g,b) ((b>>3)|((g&248)<<2)|((r&248)<<7))
#define makeCol16(r,g,b) ((b>>3)|((g&252)<<2)|((r&248)<<8))



typedef struct
{
  unsigned char r,g,b;
}rgb_type;

typedef struct
{
 char *lpbits;
 int xdim,ydim;
 int bpp;
 int errorcode;
 rgb_type palette[256];
}loaded_bitmap;


loaded_bitmap *lpBmp;
BITMAPFILEHEADER bfh;
BITMAPINFOHEADER bih30;
BITMAPCOREHEADER bih20;
HANDLE BMPfile;
char *BMPmem;
int bytes_per_pixel, bytes_per_row;
int padding;
int sizeHi;

//int bmp_flip = 1;
//int bmp_convert = 0;

struct
{
 int flip;
 int convert;
 int invert; 
 int header;
 int palette;
 int align;
}settings;


unsigned char bitfields15[12] = {0x00,0x7C,0x00,0x00,
                                 0xE0,0x03,0x00,0x00,
                                 0x1F,0x00,0x00,0x00};
unsigned char bitfields16[12] = {0x00,0xF8,0x00,0x00,
                                 0xE0,0x07,0x00,0x00,
                                 0x1F,0x00,0x00,0x00};


typedef void (*copyfunc)(char *lpsrc, char *lpdest, int xdim, int ydim, int reading);



void Copy(char *lpSrc, char *lpDest, int xdim, int ydim, int reading)
{
 int num_bytes, i;

  num_bytes = (xdim*bytes_per_pixel);

  for (i=0; i<ydim; i++)  
  {
    memcpy(lpDest, lpSrc, num_bytes);
    lpSrc += num_bytes;
    lpDest += num_bytes;
    if (reading)
      lpSrc += padding;
    else
      lpDest += padding;
  }
}


void CopyAndFlip(char *lpSrc, char *lpDest, int xdim, int ydim, int reading)
{
 int num_bytes, i;

  num_bytes = (xdim*bytes_per_pixel);

  for (i=0; i<ydim; i++)  
  {
    memcpy(lpDest, lpSrc, num_bytes);
    if (reading)
    {
      lpDest -= num_bytes;
      lpSrc += num_bytes+padding;
    }
    else
    {
      lpSrc -= num_bytes;
      lpDest += num_bytes+padding;
    }
  }
}


void BGRto555(char *lpSrc, char *lpDest, int xdim, int ydim, int reading)
{
 int num_bytes,i,j;

  num_bytes = (xdim*bytes_per_pixel);

  for (i=0; i<ydim; i++)  
  {
    for (j=0; j<xdim; j++)
    {
      *(short *)lpDest = makeCol15(*(lpSrc),*(lpSrc+1),*(lpSrc+2));
      lpSrc += 3;
      lpDest += 2;
    }
 
    if (reading)
    {
      lpSrc += padding;
    }
    else
    {
      lpDest += padding;
    }
  }
}


void ReadBMPAsBlocks(copyfunc CFunc)
{
 char *bmpBuffer, *pekare;
 int linesToRead, linesRemaining;

  bmpBuffer = LocalAlloc(LPTR, BMP_BUFFERSIZE);

  BMPmem = bmpBuffer;
  pekare = lpBmp->lpbits;

  linesToRead = (BMP_BUFFERSIZE/bytes_per_row);
  if (linesToRead>bih30.biHeight)
    linesToRead = bih30.biHeight;

  linesRemaining = bih30.biHeight;
  if (settings.flip)
    pekare += (bih30.biWidth*bytes_per_pixel*(bih30.biHeight-1));

  while (linesRemaining>0) 
  {
     ReadFile(BMPfile, bmpBuffer, (linesToRead*bytes_per_row),
              &sizeHi, NULL);
     CFunc(bmpBuffer, pekare, bih30.biWidth, linesToRead, 1);
     if (settings.flip)
       pekare -= (bih30.biWidth*bytes_per_pixel*linesToRead);
     else
       pekare += (bih30.biWidth*bytes_per_pixel*linesToRead);

     linesRemaining -= linesToRead;
     if (linesToRead>linesRemaining)
       linesToRead = linesRemaining;
  }
  LocalFree(bmpBuffer);
}


  

int ReadBMPHeader()
{
  ReadFile(BMPfile,&bfh,sizeof(BITMAPFILEHEADER),&sizeHi,NULL);
  if (bfh.bfType != 19778)
    return 1;

  ReadFile(BMPfile,&bih30,4,&sizeHi,NULL);
  if (bih30.biSize == sizeof(BITMAPINFOHEADER))
    ReadFile(BMPfile,&bih30.biWidth,sizeof(BITMAPINFOHEADER)-4,
             &sizeHi,NULL);

  else if (bih30.biSize == sizeof(BITMAPCOREHEADER))
  {
    ReadFile(BMPfile,&bih20.bcWidth,sizeof(BITMAPCOREHEADER)-4,
             &sizeHi,NULL);
    bih30.biWidth = bih20.bcWidth;
    bih30.biHeight = bih20.bcHeight;
    bih30.biPlanes = bih20.bcPlanes;
    bih30.biBitCount = bih20.bcBitCount;
  }
  else
    return 1;

  return 0;
}
  

void ReadPalette()
{
 char *palBuffer;
 int srcOffs = 0;
 int extra_byte = 0;
 int i;
 unsigned char col_mask = 0;

  if (settings.invert) col_mask = 255;

  if (bih30.biSize == sizeof(BITMAPINFOHEADER)) 
    extra_byte = 1;

  palBuffer = LocalAlloc(LPTR, 1024);
  if (extra_byte)
    ReadFile(BMPfile, palBuffer, 1024, &sizeHi, NULL);
  else
    ReadFile(BMPfile, palBuffer, 768, &sizeHi, NULL);

  for (i=0; i<256; i++)
  {
     lpBmp->palette[i].b = *(palBuffer+(srcOffs++)) ^ col_mask;
     lpBmp->palette[i].g = *(palBuffer+(srcOffs++)) ^ col_mask;
     lpBmp->palette[i].r = *(palBuffer+(srcOffs++)) ^ col_mask;
     if (extra_byte) srcOffs++;
  }
  LocalFree(palBuffer);
}
  


void Read8bit()
{
  padding = (bytes_per_row-bih30.biWidth);
  BMPmem = LocalAlloc(LPTR, (bih30.biWidth*bih30.biHeight));
  lpBmp->lpbits = BMPmem;
  lpBmp->bpp = 8;

  ReadPalette();
  bytes_per_pixel = 1;

  if (settings.flip)
    ReadBMPAsBlocks(CopyAndFlip);
  else
    ReadBMPAsBlocks(Copy);

  CloseHandle(BMPfile);
}


void Read24bit()
{
  padding = (bytes_per_row-(bih30.biWidth*3));
  BMPmem = LocalAlloc(LPTR, (bih30.biWidth*bih30.biHeight*3));
  lpBmp->lpbits = BMPmem;
  lpBmp->bpp = 24;

  bytes_per_pixel = 3;

  if (settings.flip)
    ReadBMPAsBlocks(CopyAndFlip);
  else
    ReadBMPAsBlocks(Copy);

  CloseHandle(BMPfile);
}


void LoadBMP(char *fname, loaded_bitmap *lbmp)
{
  BMPfile = CreateFile(fname, GENERIC_READ, 0, NULL, OPEN_EXISTING,
                       FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN,
                       NULL);
  if (BMPfile == INVALID_HANDLE_VALUE)
  {
    BMP_ERROR(BMP_FILE_ERROR);
  }

  lpBmp = lbmp;
  if (ReadBMPHeader())
  {
    CloseHandle(BMPfile);
    BMP_ERROR(BMP_HEADER_ERROR);
  }
  bytes_per_row = (((bih30.biBitCount*bih30.biWidth)+31)>>5)<<2;

  if (bih30.biBitCount == 8)
     if ((settings.convert == 0)||(settings.convert==8))
       Read8bit();
     else if (settings.convert == 24) return;
       //Read8bit_as_24bit();
     else
     {
       BMP_ERROR(BMP_BITCOUNT_ERROR);
     }
  else if (bih30.biBitCount == 24)
     if ((settings.convert == 0)||(settings.convert == 24))
       Read24bit();
     else
     {
       BMP_ERROR(BMP_BITCOUNT_ERROR);
     }
  else
  {
     BMP_ERROR(BMP_BITCOUNT_ERROR);
  }

  lbmp->xdim = bih30.biWidth;
  lbmp->ydim = bih30.biHeight;
  lbmp->errorcode = 0;
}





void WritePalette()
{
 char *palBuffer;
 int dstOffs = 0;
 int i;
 unsigned char col_mask = 0;

  if (settings.invert) col_mask = 255;

  palBuffer = LocalAlloc(LPTR, 1024);

  for (i=0; i<256; i++)
  {
     *(palBuffer+(dstOffs++)) = lpBmp->palette[i].b ^ col_mask;
     *(palBuffer+(dstOffs++)) = lpBmp->palette[i].g ^ col_mask;
     *(palBuffer+(dstOffs++)) = lpBmp->palette[i].r ^ col_mask;
     dstOffs++;
  }

  WriteFile(BMPfile, palBuffer, 1024, &sizeHi, NULL);
  LocalFree(palBuffer);
}


void WriteBMPAsBlocks(copyfunc CFunc)
{
 char *bmpBuffer, *pekare;
 int linesToWrite, linesRemaining;

  if (settings.align == BMP_DONT_ALIGN)
  {
    bytes_per_row -= padding;
    padding = 0;
  }
 
  bmpBuffer = LocalAlloc(LPTR, BMP_BUFFERSIZE);

  BMPmem = bmpBuffer;
  pekare = lpBmp->lpbits;

  linesToWrite = (BMP_BUFFERSIZE/bytes_per_row);
  if (linesToWrite>bih30.biHeight)
    linesToWrite = bih30.biHeight;

  linesRemaining = bih30.biHeight;
  if (settings.flip)
    pekare += (bih30.biWidth*bytes_per_pixel*(bih30.biHeight-1));

  while (linesRemaining>0) 
  {
     CFunc(pekare, bmpBuffer, bih30.biWidth, linesToWrite, 0);
     WriteFile(BMPfile, bmpBuffer, (linesToWrite*bytes_per_row),
              &sizeHi, NULL);
     if (settings.flip)
       pekare -= (bih30.biWidth*bytes_per_pixel*linesToWrite);
     else
       pekare += (bih30.biWidth*bytes_per_pixel*linesToWrite);

     linesRemaining -= linesToWrite;
     if (linesToWrite>linesRemaining)
       linesToWrite = linesRemaining;
  }
  LocalFree(bmpBuffer);
}


void Write8bit()
{
  padding = (bytes_per_row-bih30.biWidth);
  bytes_per_pixel = 1;
  if (settings.flip)
    WriteBMPAsBlocks(CopyAndFlip);
  else
    WriteBMPAsBlocks(Copy);
  CloseHandle(BMPfile);
}


void Write24bitAs15bit()
{
  padding = (bytes_per_row-bih30.biWidth);
  bytes_per_pixel = 3;
  if (settings.flip)
    WriteBMPAsBlocks(CopyAndFlip);
  else
    WriteBMPAsBlocks(BGRto555);
  CloseHandle(BMPfile);
}


int SaveBMP(char *fname, loaded_bitmap *srcbmp, loaded_bitmap *trgtbmp)
{

  BMPfile = CreateFile(fname, GENERIC_WRITE,0,NULL,CREATE_ALWAYS,
                       FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN,NULL);
  if (BMPfile == INVALID_HANDLE_VALUE)
    return BMP_FILE_ERROR;

  lpBmp = srcbmp;

  bfh.bfType = 19778;
  bih30.biSize = sizeof(BITMAPINFOHEADER);
  bih30.biWidth = trgtbmp->xdim;
  bih30.biHeight = trgtbmp->ydim;
  bih30.biBitCount = trgtbmp->bpp;
  bih30.biSizeImage = (bih30.biWidth*bih30.biHeight)*((bih30.biBitCount+1)>>3);
  bih30.biPlanes = 1;
  if ((bih30.biBitCount==8)||(bih30.biBitCount==24))
    bih30.biCompression = BI_RGB;
  else
    bih30.biCompression = BI_BITFIELDS;
  bih30.biXPelsPerMeter = 72;
  bih30.biYPelsPerMeter = 72;
  bih30.biClrUsed = 0;
  bih30.biClrImportant = 0;

  bfh.bfSize = bih30.biSizeImage+54;
  bfh.bfReserved1 = 0;
  bfh.bfReserved2 = 0;
  bfh.bfOffBits = 40;

  bytes_per_row = (((bih30.biWidth*bih30.biBitCount)+31)>>5)<<2;

  if (bih30.biBitCount == 8) 
  {
    bfh.bfSize += 1024;
    bfh.bfOffBits += 1024;
  }
  else if (bih30.biBitCount == 15) 
  {
    bih30.biBitCount = 16;
    bfh.bfSize += 12;
    bfh.bfOffBits += 12;
  }
  else if (bih30.biBitCount == 16)
  {
    bfh.bfSize += 12;
    bfh.bfOffBits += 12;
  }

  if (settings.header)
  {
    WriteFile(BMPfile,&bfh,sizeof(BITMAPFILEHEADER),&sizeHi,NULL);
    WriteFile(BMPfile,&bih30,sizeof(BITMAPINFOHEADER),&sizeHi,NULL);
  }

  switch (bih30.biBitCount)
  {
    case 8:
      if (srcbmp->bpp == 8)
      {
        if (settings.palette) WritePalette();
        Write8bit();
      }
      else
        return 1;
      break;

    case 15:
//      WriteFile(BMPfile,&bitfields15[0],12,&sizeHi,NULL);
      if (srcbmp->bpp == 24)
        Write24bitAs15bit();
      else
        return 1;
      break;

    case 24:
      break;

    default:
      break;
  }
  return 0;
}



void ConfigBMP(int pname, int pvalue)
{
  switch (pname)
  {
    case BMP_FLIP:
      settings.flip = pvalue;
      break;
    case BMP_INVERT:
      settings.invert = pvalue;
      break;
    case BMP_HEADER:
      settings.header = pvalue;
      break;
    case BMP_PALETTE:
      settings.palette = pvalue;
      break;
    case BMP_ALIGN:
      settings.align = pvalue;
      break;
    default:
      break;
  }
}  