/*

  Routines for reading & writing GIF & PCX images
  Can also save as Targa (.tga), although only as 8-bit

  (based on code by Michael Bolin)

  /Mic, 2001

*/


#include "beeampee4.h"
#include <math.h>


#define	PCX_OPEN_FAILED		1
#define	PCX_UNEXPECTED_EOF	2
#define	PCX_UNSUPPORTED_FORMAT	3
#define	GIF_OPEN_FAILED		1
#define	GIF_UNEXPECTED_EOF	2
#define GIF_UNSUPPORTED_FORMAT	3

#define BIT0	1
#define BIT1	2
#define BIT2	4
#define BIT3	8
#define BIT4	16
#define BIT5	32
#define BIT6	64
#define BIT7	128


#define hash 11003

/* "GIF" */
#define GIFSIG	4606279

typedef struct
{
 char signature[3];
 char version[3];
 unsigned short logical_screen_width;
 unsigned short logical_screen_height;
 unsigned char packed_fields;
 unsigned char background_color;
 unsigned char pixel_aspect_ratio;
}GIFHEADER;

typedef struct
{
 char separator;
 unsigned short left_position;
 unsigned short top_position;
 unsigned short width;
 unsigned short height;
 unsigned char packed_fields;
}GIFDESCRIPTOR;


typedef struct
{
 char manufacturer;
 char version;
 char encoding;
 char bitsperpixel;
 short xmin,ymin,xmax,ymax;
 short hdpi,vdpi;
 char colormap[48];
 char reserved;
 char nplanes;
 short bytesperline;
 short paletteinfo;
 short hscreensize;
 short vscreensize;
 char filler[54];
}PCXHEADER;


typedef loaded_bitmap GIFIMAGE;
typedef loaded_bitmap PCXIMAGE;
typedef loaded_bitmap TGAIMAGE;


HANDLE hGif,hPcx,hTga;


int read_gif_header(GIFIMAGE *pGIF)
{
 int i,j;
 int bytesRead;
 int color_table,color_resolution,sort_flag,sizeof_color_table;
 int n_colors;
 char c1;
 static GIFHEADER gh;
 

  ReadFile(hGif, &gh, sizeof(GIFHEADER), &bytesRead, NULL);
  if (bytesRead<sizeof(GIFHEADER))
    return GIF_UNEXPECTED_EOF;

  if ((gh.signature[0]!='G')||(gh.signature[1]!='I')||
      (gh.signature[2]!='F'))
    return GIF_UNSUPPORTED_FORMAT;

  color_table = (gh.packed_fields & BIT7);
  sizeof_color_table = (gh.packed_fields & 7)+1;
  n_colors = pow(2,sizeof_color_table);

  if (color_table)
  for (i=0; i<n_colors; i++)
    ReadFile(hGif, &pGIF->palette[i], 3, &bytesRead, NULL);

  return 0;
}


int read_gif_image(GIFIMAGE *pGIF)
{
 int i,j;
 int bytesRead;
 int color_table,interlace,sort_flag,sizeof_color_table;
 int n_colors;
 int clear_code,end_of_info,num_bits;
 unsigned char code_size,block_length;
 static char gif_stream[256]; 
 static GIFDESCRIPTOR gd;

  ReadFile(hGif, &gd, sizeof(GIFDESCRIPTOR), &bytesRead, NULL);
  if (bytesRead<sizeof(GIFDESCRIPTOR))
    return GIF_UNEXPECTED_EOF;

  color_table = (gd.packed_fields & BIT7);
  interlace = (gd.packed_fields & BIT6);
  sizeof_color_table = (gd.packed_fields & 7)+1;
  n_colors = pow(2, sizeof_color_table);

  if (color_table)
  for (i=0; i<n_colors; i++)
    ReadFile(hGif, &pGIF->palette[i], 3, &bytesRead, NULL);

  pGIF->xdim = gd.width;
  pGIF->ydim = gd.height;

  if (n_colors==2)
    n_colors = 4;

  ReadFile(hGif, &code_size, 1, &bytesRead, NULL);
  clear_code = pow(2,code_size);
  end_of_info = clear_code+1;
  num_bits = code_size+1;

  i = 0;
  while (1)
  {
    ReadFile(hGif, &block_length, 1, &bytesRead, NULL);
    if (block_length==0)
       break;
    for (j=0; j<block_length; j++)
       ReadFile(hGif, &gif_stream[i++], 1, &bytesRead, NULL);
  }

// !!!!! Unfinished  (gifpcx.e, row 380) !!!!!

return 0;
}



int read_pcx_header(PCXIMAGE *pPCX)
{
 int i;
 int bytesRead;
 int palette_at_end;
 static PCXHEADER ph;

  ReadFile(hPcx, &ph, sizeof(PCXHEADER), &bytesRead, NULL);
  if (bytesRead<sizeof(PCXHEADER))
    return PCX_UNEXPECTED_EOF;

  if (ph.encoding != 1)
    return PCX_UNSUPPORTED_FORMAT;

  palette_at_end = 0;
  if (ph.version == 5)
    palette_at_end = 1;

  if (ph.bitsperpixel==1)
    palette_at_end = 0;

  if (!palette_at_end)
    for (i=0; i<16; i++)
    {
      pPCX->palette[i].r = ph.colormap[(i<<1)+i];
      pPCX->palette[i].g = ph.colormap[(i<<1)+i+1];
      pPCX->palette[i].b = ph.colormap[(i<<1)+i+2];
    }

  return 0;
}

  

void LoadGIF(char *fname, GIFIMAGE *pGIF)
{
}

void LoadPCX(char *fname, PCXIMAGE *pPCX)
{
}



int SaveGIF(char *fname, GIFIMAGE *pGIF)
{
  return 0;
}


int SavePCX(char *fname, PCXIMAGE *pPCX)
{
  return 0;
}


int SaveTGA(char *fname, TGAIMAGE *pTGA)
{
  return 0;
}
