/*

 Various routines for ordered lists and sorting
 /Mic, 2001

*/


#include <windows.h>

/* Indexing macros for the children of a node */
#define LC(n)		(n<<1)
#define RC(n)		(n<<1)+1

typedef struct
{
 int id;
 int count;
}item;


typedef struct
{
 item *data;
 int size;
}Heap;



void CreateHeap(Heap *H, int numitems)
{
  H->data = LocalAlloc(LPTR, sizeof(item)*numitems);
  H->size = numitems;
}

void DestroyHeap(Heap *H)
{
  LocalFree(H->data);
}


void siftdown(Heap *H, int L, int U)
{
 int L2 = L;
 item temp;
 int LCL, RCL;

  while (1)
  {
    LCL = LC(L2);
    RCL = RC(L2);

    if ( (RCL <= U) && (H->data[RCL-1].count >= H->data[LCL-1].count) &&
         (H->data[RCL-1].count > H->data[L2-1].count) )
    {
       temp = H->data[L2-1];
       H->data[L2-1] = H->data[RCL-1];
       H->data[RCL-1] = temp;
       L2 = RCL;
    }
    else if ( (LCL <= U) && (H->data[LCL-1].count > H->data[L2-1].count) )
    {
       temp = H->data[L2-1];
       H->data[L2-1] = H->data[LCL-1];
       H->data[LCL-1] = temp;
       L2 = LCL;
    }
    else
       return;
  }
}



// Search for an item with ID 'key'.
// Exploit the ordering by doing a binary search.
// 
int OrderedLookup(int key, Heap *H, int items)
{
 int L,U,middle;

  L = 0;
  U = items-1;

  while (1)
  {
  if (L>U)
    break;

  middle = (L+U)>>1;
  if (key == H->data[middle].id)
    return middle;
  else if (key < H->data[middle].id)
    U = middle-1;
  else
    L = middle+1;
  }
  return -1;
}


// Insert 'I' at position 'items'.
// Items [0..'items'-1] are assumed to be in ascending order.
// After the call, items [0..'items'] will be in ascending order. 
//
void OrderedInsert(item *I, Heap *H, int items)
{
 int j;
 item temp;

  H->data[items] = *I;
  j = items;

  /* Make sure the new item ends up in the right position */
  while (j-1>=0)
  {
     if (H->data[j].id < H->data[j-1].id)  
     {
       temp = H->data[j];
       H->data[j] = H->data[j-1]; 
       H->data[j-1] = temp;
     }
     j--;
  }
} 



// Heapsort heap H.
//
void HeapSort(Heap *H, int steps)
{
 int i;
 item temp;
 int half = H->size>>1;

  if (!(H->size & 1)) half--;
  for (i=half; i>=1; i--)
    siftdown(H, i, H->size);

  for (i=1; i<=steps; i++)
  {
    temp = H->data[0];
    H->data[0] = H->data[H->size-i];
    H->data[H->size-i] = temp;
    siftdown(H, 1, H->size-i);
  }
}


// Shellsort heap H.
// (generally slower than the above heapsort).
//
void ShellSort(Heap *H)
{
 int i,j,gap,last;
 item tempi,tempj;

  last = H->size-1;
  gap = ((H->size)/3)+1;

  while (1)
  {
    for (i=gap; i<=last; i++)
    {
      tempi = H->data[i];
      j = i-gap;

      while (1)
      {
        tempj = H->data[j];
        if (tempj.count<tempi.count)
          { j += gap; break; }
        H->data[j+gap] = tempj;
        if (j<=gap)
          break;
        j -= gap;
      }
      H->data[j] = tempi;
    }
    if (gap==1) return;
    gap = (gap/3)+1;
  }
}




